import argparse
import os
from six.moves import cStringIO as StringIO

import numpy as np
import pandas as pd
import scipy.stats as stats
from tornado import ioloop, web
from projectile.server import CmapTileHandler

try:
    import pyBigWig
    bigwig_avail = True
except ImportError:
    bigwig_avail = False

from hiclite.util.sparse_matrix import SparseContactMatrix

import hiczoom


PATH = os.path.abspath(os.path.dirname(__file__))


class TsvChromHandler(web.RequestHandler):
    def initialize(self, df, filter_column='chrom'):
        self.df = df
        self.filter_column = filter_column

    def get(self, chrom):
        stream = StringIO()
        self.df[self.df[self.filter_column] == chrom].to_csv(
            stream, sep='\t', index=False)
        for line in stream.getvalue():
            self.write(line)
        self.set_header('Content-type', 'text/tab-separated-values')


class ValueHandler(web.RequestHandler):
    def initialize(self, value):
        self.value = value

    def get(self):
        self.write(str(self.value))
        self.set_header('Content-type', 'text/plain')


class ClientHandler(web.RequestHandler):
    def initialize(self, client, chrom, cmap='Reds', vmin=0, vmax=25,
                   tile_size=256):
        self.client = client
        self.chrom = chrom
        self.cmap = cmap
        self.vmin = vmin
        self.vmax = vmax
        self.tile_size = tile_size

    def get(self):
        self.render(self.client, chrom=self.chrom, cmap=self.cmap,
                    vmin=self.vmin, vmax=self.vmax, tile_size=self.tile_size)


class BigWigHandler(web.RequestHandler):
    def initialize(self, bw, chrom_sizes, resolution):
        self.bw = bw
        self.chrom_sizes = chrom_sizes
        self.resolution = resolution

    def get(self, c, z, x):
        z, x = map(int, (z, x))
        chrom_size = self.chrom_sizes[self.chrom_sizes['chrom'] == c]['size']
        max_z = int(np.ceil(np.log2(chrom_size / float(self.resolution))))
        size = (2 ** (max_z - z)) * self.resolution
        if self.bw:
            values = self.bw.stats(c, x * size, (x+1) * size, type='max',
                                   nBins=500)
        else:
            values = np.zeros(500)
        coords = np.linspace(x * size, (x+1) * size, num=500, endpoint=False) \
            + size / 1000
        stream = StringIO()
        np.savetxt(stream, np.array([coords, values]).T, delimiter='\t',
                   fmt='%s')
        self.write('coord\tvalue\n')
        for line in stream.getvalue():
            self.write(line)
        self.set_header('Content-type', 'text/tab-separated-values')


def make_app(array, chrom_sizes, resolution, chrom, tile_size=256, client=None,
             loops=None, genes=None, bigwig=None, cmap='Reds', vmin=0, vmax=25,
             debug=False):
    if client is None:
        client = '%s/client.html' % PATH
    return web.Application([
        (r'/', ClientHandler,
         {'client': client, 'chrom': chrom, 'cmap': cmap, 'vmin': vmin,
          'vmax': vmax, 'tile_size': tile_size}),
        (r'/resolution/', ValueHandler, {'value': resolution}),
        (r'/loops/(\w+).tsv', TsvChromHandler,
         {'df': loops, 'filter_column': 'us_chrom'}),
        (r'/genes/(\w+).tsv', TsvChromHandler, {'df': genes}),
        (r'/chrom_sizes/(\w+).tsv', TsvChromHandler, {'df': chrom_sizes}),
        (r'/chipseq/(\w+)/([0-9]+)/([0-9]+).tsv', BigWigHandler,
         {'bw': bigwig, 'chrom_sizes': chrom_sizes, 'resolution': resolution}),
        (r'/([0-9]+)/([0-9]+)/([0-9]+)/([0-9]+)/(\w+)/([^/]+)/([^/]+).png',
         CmapTileHandler, {'array': array})
    ], debug=debug)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'matrix', help='''Contact matrix to visualize. Should be either a sparse
        .npz file or a dense .npy file.''')
    parser.add_argument(
        'genome', choices=['hg18', 'hg19', 'hg38', 'mm9', 'mm10'],
        help='''Genome assembly to use.''')
    parser.add_argument(
        'chrom', help='''Name of the chromosome corresponding to the matrix.''')
    parser.add_argument(
        'resolution', type=int, help='''Resolution of the contact matrix in
        bp.''')
    parser.add_argument(
        '-c', '--cmap', default='Reds', help='''Name of the matplotlib colormap
        used to color the contact matrix. Default is "Reds".''')
    parser.add_argument(
        '-Z', '--zscore', action='store_true', help='''If matrix contains
        p-values, pass this flag to convert them to z-scores using the probit
        function.''')
    parser.add_argument(
        '-l', '--loops', help='''Pass a .tsv file of called loops (hiclite
        format) to outline them with green boxes.''')
    parser.add_argument(
        '-b', '--bigwig', help='''Pass a .bw file to visualize it along the axes
        of the contact matrix. Requires pyBigWig to be installed.''')
    parser.add_argument(
        '-m', '--max_range', type=int, default=300, help='''If matrix is dense,
        specify how many off-diagonals to keep when converting to the sparse
        format. Default is 300, corresponding to 3 Mb at 10 kb resolution.''')
    parser.add_argument(
        '--vmax', type=float, default=25, help='''Specify the initial vmax of
        the colorscale. Default is 25.''')
    parser.add_argument(
        '--vmin', type=float, default=0, help='''Specify the initial vmin of
        the colorscale. Default is 0.''')
    parser.add_argument(
        '-t', '--tile_size', type=int, default=256, help='''The image tile size
        to use (in pixels). Must be a power of 2. Default is 256 for 256 x 256
        image tiles.''')
    parser.add_argument(
        '--client', help='''Specify a custom client HTML file to serve.''')
    parser.add_argument(
        '-p', '--port', type=int, default=8000, help='''The port to start the
            server on. Default is 8000.''')
    parser.add_argument(
        '-D', '--debug', action='store_true', help='''Pass this flag to start
            the server in debug mode.''')
    parser.add_argument('-v', '--version', action='version',
                        version='hiczoom version %s' % hiczoom.__version__)
    args = parser.parse_args()

    print('loading data')
    scm = SparseContactMatrix.load(args.matrix, max_range=args.max_range)
    if args.zscore:
        print('converting to z-scores')
        scm.data = stats.norm.isf(scm.data)
    scm.symmetrize()
    genes = pd.read_csv(
        '%s/gene_tracks/%s_refseq_genes.gz' % (PATH, args.genome), sep='\t')
    loops_file = '%s/blank_loops.tsv' % PATH if args.loops is None \
        else args.loops
    loops = pd.read_csv(loops_file, sep='\t')
    chrom_sizes = pd.read_csv(
        '%s/chrom_sizes/%s.chrom.sizes' % (PATH, args.genome), sep='\t',
        names=['chrom', 'size'])
    bw = None if args.bigwig is None or not bigwig_avail \
        else pyBigWig.open(args.bigwig)

    print('starting server at http://localhost:%s/' % args.port)
    app = make_app(
        scm,
        chrom_sizes,
        args.resolution,
        args.chrom,
        tile_size=args.tile_size,
        client=args.client,
        loops=loops,
        genes=genes,
        bigwig=bw,
        cmap=args.cmap,
        vmin=args.vmin,
        vmax=args.vmax,
        debug=args.debug
    )
    app.listen(args.port)
    ioloop.IOLoop.instance().start()


if __name__ == '__main__':
    main()
